function [ U, Sigma, V ] = BwzFixedRankApprox( obj, r )
%BWZFIXEDRANKAPPROX Boutisidis' Single-View Fixed-Rank Approximation.
%This function implements Boutisidis' Single-View Fixed-Rank Approximation
%[BWZ2016], using simplifications suggested in [Upa2016]. This method is
%explained in Section 7.3.2 in our main reference [TYUC2017], see (7.6).
%   Require: Target rank (r <= k)
%   Ensure:  Returns factors 
%             (m x r) dimensional Q with orthonormal columns
%             (n x r) dimensional V with orthonormal columns
%             (r x r) dimensional nonnegative diagonal matrix (Sigma) 
%            that form a rank-r approximation (Abwz = U*Sigma*V') of the 
%            sketched matrix.
%
%   Ar = S.BwzFixedRankApprox(r) returns the rank-r approximation of the
%   target matrix A. 
%
%   [U, Sigma, V] = S.BwzFixedRankApprox(r) returns the factors U, Sigma
%   and V that form the rank-r apptroximation Abwz = U*Sigma*V'. Here, U
%   and V are (m x r) and (n x r) dimensional factors with orthonormal
%   columns, and Sigma is the (r x r) dimensional matrix. 
%
%See our reference paper [TYUC2017] for the detailed explanation of the
%sketching procedure and the arithmetic and storage costs.
%
%[BWZ2016] C. Boutsidis, D. Woodruff and P. Zhong, Optimal Principal 
%Component Analysis in Distributed and Streaming Models, 2016.
%
%[Upa2016] J. Upadhyay, Fast and Space-Optimal Low-Rank Factorization in 
%the Streaming Model with Application in Differential Privacy, 2016.
%
%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical
%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%Coded by: Alp Yurtsever
%Ecole Polytechnique Federale de Lausanne, Switzerland.
%Laboratory for Information and Inference Systems, LIONS.
%contact: alp.yurtsever@epfl.ch
%Created: December 30, 2016
%Last modified: August 26, 2016
%
%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox. 
%Please read COPYRIGHT before using this file.

if r > min(size(obj.Omega, 1), size(obj.Psi,2))
    error('Target rank ''r'' cannot be greater than matrix dimensions.')
end

% Computing the approximation
[Q1, ~]   = qr(obj.Y,0);
[Q2, ~]   = qr(obj.W',0); 
[U1, T1]  = qr(obj.Phi(Q1),0);
[U2, T2]  = qr(obj.Xi(Q2')',0);
[Uupa, SigmaUpa, Vupa] = svd(U1'*obj.Z*U2);

U = Q1*(T1 \ Uupa(:,1:r));
Sigma = SigmaUpa(1:r,1:r);
V = Q2*(Vupa(:,1:r)' / T2')';

if nargout == 1; U = U*Sigma*V'; end

end

