function obj = LinearUpdate( obj, varargin )
%LINEARUPDATE Linear Update method for OptSketch class
%This function implements a linear update routine for the more complicated 
%sketches described in Section 7.3.2 of [TYUC2017]. 
%   Require: (m x n) dimensinal update matrix (H)
%            scalars (theta) and (eta)
%   Ensure:  Modifies sketch (Y,W) to reflect linear update
%                    A = theta*A + eta*H
%            which takes the form
%                    Y = theta*Y + eta*H*Omega
%                    W = theta*W + eta*Psi*H
%                    Z = theta*W + eta*Phi*H*Xi
%
%   S = S.LinearUpdate(H, theta, eta) updates the sketches Y and W as 
%   Y = theta*Y + eta*H*Omega and W = theta*W + eta*Psi*H.
%
%   S = S.LinearUpdate(U, V, theta, eta) where Hf and Hb are tall
%   matrices updates the sketches Y and W from the factors H = U*V' 
%   in a storage efficient way.
%
%See our reference paper [TYUC2017] for the detailed explanation of the
%sketching procedure and the arithmetic and storage costs.
%
%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical
%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%Coded by: Alp Yurtsever
%Ecole Polytechnique Federale de Lausanne, Switzerland.
%Laboratory for Information and Inference Systems, LIONS.
%contact: alp.yurtsever@epfl.ch
%Created: December 30, 2016
%Last modified: January 13, 2017
%
%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox. 
%Please read COPYRIGHT before using this file.

narginchk(4,5);
if nargin == 4
    H       = varargin{1};
    theta   = varargin{2};
    eta     = varargin{3};
    obj.Y = theta*obj.Y + eta*(H*obj.Omega);
    obj.W = theta*obj.W + eta*(obj.Psi*H);
    obj.Z = theta*obj.Z + eta*(obj.Phi(obj.Xi(H)));
else
    U   = varargin{1};
    V   = varargin{2};
    theta   = varargin{3};
    eta     = varargin{4};
    obj.Y = theta*obj.Y + eta*(U*(V'*obj.Omega));
    obj.W = theta*obj.W + eta*((obj.Psi*U)*V');
    obj.Z = theta*obj.Z + eta*(obj.Phi(U)*obj.Xi(V'));
end

end

