classdef OptSketch
    %OPTSKETCH implements a class definition for the more complicated 
    %sketching methods described in Section 7.3.2 of [TYUC2017]. We 
    %implement two methods for this class definition. 
    %
    %   List of methods:
    %   obj = LinearUpdate(obj, varargin) 
    %   [Q, Z, V] = BwzFixedRankApprox(obj, r)  	% Algorithm described in eqn. (7.6) in [TYUC2017]
	%
    %See our reference paper [TYUC2017] for the detailed explanation of the
    %sketching procedure and the arithmetic and storage costs.
    %
	%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
	%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
    %
    %Coded by: Alp Yurtsever
    %Ecole Polytechnique Federale de Lausanne, Switzerland.
    %Laboratory for Information and Inference Systems, LIONS.
    %contact: alp.yurtsever@epfl.ch
    %Created: December 28, 2016
    %Last modified: August 26, 2017
    %
    %PRACTICALSKETCHING-v1.0
    %Copyright (C) 2017 Laboratory for Information and Inference Systems
    %(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
    %
    %This code is a part of PRACTICALSKETCHING toolbox.
    %Please read COPYRIGHT before using this file.
    
    %% properties
    properties
        Field     % 'real' or 'complex' field (default, 'complex')
        Omega     % (n x k) dimensional test matrix for the range of A (std Gaussian)
        Psi       % (k x m) dimensional test matrix for the corange of A (std Gaussian)
        Phi       % (s x m) dimensional test matrix for the inner sketch (SRFT)
        Xi        % (n x s) dimensional test matrix for the inner sketch (SRFT)
        Y         % (m x k) dimensional range sketch
        W         % (k x n) dimensional corange sketch
        Z         % (s x s) dimensional inner sketch
    end
    
    %% methods
    methods
        % Constructor
        % function obj = OptSketch(A, k, s, Field, Orthogonalization)
        function obj = OptSketch(A, k, s, varargin)
            % process variable length inputs
            % 1. Field, 2. Orthogonalization
            narginchk(3,5);
            if nargin >= 4
                obj.Field  = lower(varargin{1});
            else
                obj.Field = 'complex';
            end
            if nargin == 5
                Orthogonalization = varargin{2};
            else
                Orthogonalization = 0;
            end
            % construct the sketch
            [m, n] = size(A);
            if (m + n) == 3     % if A is 2 dimensional, consider it as the dimensions [m,n]
                m = A(1);
                n = A(2);
                A = sparse(m,n);
            end
            if strcmp(obj.Field,'real')
                obj.Omega = randn(n,k);
                obj.Psi   = randn(k,m);
            elseif strcmp(obj.Field,'complex')
                obj.Omega = randn(n,k) + 1i*randn(n,k);
                obj.Psi   = randn(k,m) + 1i*randn(k,m);
            else
                error('Field should be ''real'' or ''complex''.');
            end
            if Orthogonalization
                obj.Omega = orth(obj.Omega);
                obj.Psi   = orth(obj.Psi')';
            end
            % create test matrices for inner sketch (SRFT, see section 3.9 for details)
            % NOTE: below lines assume that the field is complex! For real
            % field, discrete Fourier transform needs to be replaced by
            % discrete cosine transform following some simple and
            % straightforward modifications.
            ssampleIdx = @(x,idx) x(idx,:);
            d1 = sign(randn(n,1));
            idx1 = sort(randperm(length(d1),s));
            obj.Phi = @(x) ssampleIdx( fft(x.*repmat(d1,[1,size(x,2)]), [], 1), idx1);
            d2 = sign(randn(n,1));
            idx2 = sort(randperm(length(d2),s));
            obj.Xi = @(x) ssampleIdx( fft(x'.*repmat(d2,[1,size(x',2)]), [], 1), idx2)';
            % initialize the sketch
            obj.Y = A*obj.Omega;
            obj.W = obj.Psi*A;
            if issparse(A)
                if nnz(A)
                    obj.Z = obj.Phi(obj.Xi(full(A)));
                else
                    obj.Z = zeros(s,s);
                end
            else
                obj.Z = obj.Phi(obj.Xi(A));
            end
        end
        
        % Other methods
        obj = LinearUpdate(obj, varargin)           
        [Q, Sigma, V] = BwzFixedRankApprox(obj,r)
    end
end

% TBD: Add other distributions for the test matrices!
