function [ V, Z ] = CemmpFixedRankApprox( obj, r )
%CEMMPFIXEDRANKAPPROX Single-View Fixed-Rank Approximation by Cohen et al.
%This function implements the method implicitly defined in [CEMMP2015, Sec.10.1].
%This method takes the form (7.3) in the main reference [TYUC2017].
%   Require: Target rank (r <= k)
%   Ensure:  Returns the low rank reconstrution Acoh of the sketched
%            matrix.
%
%   Acoh = S.CemmpFixedRankApprox(r) the rank-r approximation of the target
%   matrix A. This implements Cohen et al.'s fixed rank approximation
%   formuluation (7.3) in [TYUC2017].
%
%   [V, Z] = S.CemmpFixedRankApprox(r) returns the factors V and Z
%   that forms Acoh = V*Z, where Acoh is the output of the Cohen et al.'s
%   fixed rank approximation formuluation (7.3) in [TYUC2017].
%
%See our reference paper [TYUC2017] for the detailed explanation of the 
%sketching procedure and the arithmetic, communication and storage costs.
%
%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%[CEMMP2015] M.B. Cohen, S. Elder, C. Musso, C.Musso, and M. Persu, 
%Dimensionality Reduction for k-means Clustering and Low-Rank
%Approximation, in Proceedings of the 47th Annual ACM on Symposium on
%Theory of Computing, pp. 163-172, 2015.
%
%Coded by: Alp Yurtsever
%Ecole Polytechnique Federale de Lausanne, Switzerland.
%Laboratory for Information and Inference Systems, LIONS.
%contact: alp.yurtsever@epfl.ch
%Created: December 30, 2016
%Last modified: September 2, 2017
%
%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox. 
%Please read COPYRIGHT before using this file.

if r > min(size(obj.Omega, 1), size(obj.Psi,2))
    error('Target rank ''r'' cannot be greater than matrix dimensions.')
end

[V,~,~] = svd(obj.Y,0);
V       = V(:,1:r);
[U,T]   = qr(obj.Psi*V,0);

X = U'*obj.W;
if min(size(X)) > 5000
    [Ucoh2, Sigma2, Vcoh2] = svds(X,r);
else
    [Ucoh2, Sigma2, Vcoh2] = svd(X,'econ');
    Ucoh2 = Ucoh2(:,1:r);
    Sigma2 = Sigma2(1:r,1:r);
    Vcoh2 = Vcoh2(:,1:r);
end
Z = T\(Ucoh2*(Sigma2*Vcoh2'));

if nargout == 1; V = V*Z; end

end

