classdef Sketch
    %SKETCH of a (m x n) dimensional target matrix A takes the form
    %(Y = A*Omega) and (W = Psi*A) which are (m x k) and (l x n)
    %dimensional respectively. Here (n x k) dimensional matrix (Omega) and
    %(l x m) dimensional matrix (Psi) are called as the Test Matrices. This
    %code implements the sketching method proposed in the reference paper,
    %where the test matrices are statistically independent and follow the
    %standard normal distribution. The matrix (Y) collects information
    %about the range of (A), while the matrix (W) collects information
    %about the co-range of (A). Both parts are necessary.
    %
    %   S = Sketch(A, k ,l) returns the sketch object consisting Gaussian 
    %   test matrices (n x k) dimensional 'Omega' and (l x m) dimensional 
    %   'Psi', as well as the sketches 'Y' and 'W' of the target matrix A. 
    %
    %   S = Sketch([m, n], k ,l) returns the sketch object consisting  
    %   Gaussian test matrices (n x k) dimensional 'Omega' and (l x m)  
    %   dimensional 'Psi', and initialize the sketches 'Y' and 'W' for the
    %   zero target matrix A. 
    %
    %   S = Sketch(A, k, l, Field) returns the sketch object in the field 
    %   specified, which can be 'real' or 'complex' (default: 'complex').
    %
    %   S = Sketch(A, k, l, Field, 1) returns the sketch object in the 
    %   field specified and with orthogonal test matrices.
    %
    %   List of methods:
    %   obj = LinearUpdate(obj, varargin)           % Algorithm 2 in [TYUC2017]
    %   [Q, X] = SimpleLowRankApprox(obj)           % Algorithm 3 in [TYUC2017]
    %   [Q, X] = LowRankApprox(obj)                 % Algorithm 4 in [TYUC2017]
    %   [U, S] = LowRankSymApprox(obj)              % Algorithm 5 in [TYUC2017]
    %   [U, D] = LowRankPSDApprox(obj)              % Algorithm 6 in [TYUC2017]
    %   [Q, Sigma, V] = FixedRankApprox(obj, r)     % Algorithm 7 in [TYUC2017]
    %   [U, D] = FixedRankSymApprox(obj, r)         % Algorithm 8 in [TYUC2017]
    %   [U, D] = FixedRankPSDApprox(obj, r)         % Algorithm 9 in [TYUC2017]
    %   [Awoo] = WoodruffLowRankApprox(obj)         % Woodruff's low rank approx (algebraically equivalent to Algorithm 4, see remark 4.1 in [TYUC2017])
    %   [Q, Z] = WoodruffFixedRankApprox(obj, r)    % Woodruff's fixed rank approx (7.2) in [TYUC2017]
    %   [V, Z] = CemmpFixedRankApprox(obj, r)       % Cohen et al.'s fixed rank approx (7.3) in [TYUC2017]
    %
    %   S = S.LinearUpdate(H, theta, eta) updates the sketches Y and W as
    %   Y = theta*Y + eta*H*Omega and W = theta*W + eta*Psi*H.
    %
    %   S = S.LinearUpdate(Hf, Hb, theta, eta) where Hf and Hb are tall
    %   matrices updates the sketches Y and W from the factors H = Hf*Hb'
    %   in a storage efficient way.
    %
    %   S = S.LinearUpdate(Hforw, Hback, theta, eta) where Hforw and Hback
    %   are the function handles to compute H*x = Hforw(x) and x*H = Hback(x),
    %   updates the sketches Y and W as Y = theta*Y + eta*H*Omega and
    %   W = theta*W + eta*Psi*H.
    %
    %   Aout = S.SimpleLowRankApprox() returns the rank-q approximation of
    %   the target matrix A for some q <= k. This implements Algorithm 3.
    %
    %   [Q, X] = S.SimpleLowRankApprox() returns the factors Q and X that
    %   form the rank-q approximation Aout = Q*X for some q <= k. Here, Q 
    %   is the (m x q) dimensional factor with orthonormal columns, and X is
    %   of (q x n) dimensions. This implements Algorithm 3.
    %   
    %   Aout = S.LowRankApprox() returns the rank-q approximation of the target 
    %   matrix A for some q <= k. This implements Algorithm 4.
    %
    %   [Q, X] = S.LowRankApprox() returns the factors Q and X that
    %   form the rank-q approximation Aout = Q*X for some q <= k. Here, Q 
    %   is the (m x q) dimensional factor with orthonormal columns, and X 
    %   is of (q x n) dimensions. This implements Algorithm 4.
    %
    %   Aout = S.FixedRankApprox(r) returns the rank-r approximation of the
    %   target matrix A. This implements Algorithm 5.
    %
    %   [Q, Sigma, V] = S.FixedRankApprox(r) returns the factors Q, Sigma
    %   and V that form the rank-r apptroximation Aout = U*Sigma*V'. Here, 
    %   U and V are (m x r) and (n x r) dimensional factors with  
    %   orthonormal columns, and Sigma is the (r x r) dimensional  
    %   nonnegative diagonal matrix. This implements Algorithm 5.
    %
    %   Aout = S.LowRankSymApprox() returns the rank-q conjugate symmetric 
    %   approximation of the square target matrix A for some q <= k. This
    %   implements Algorithm 6. 
    %
    %   [U, D] = S.LowRankSymApprox() returns the factors U and D that
    %   form the rank-q conjugate symmetric approximation Aout = U*D*U' for  
    %   some q <= k. Here, U  is the (m x q) dimensional factor with  
    %   orthonormal columns, and S is of (q x q) dimensions and conjugate  
    %   symmetric. This implements Algorithm 6.
    %
    %   Aout = S.LowRankPSDApprox() returns the rank-q positive  
    %   semi-definite approximation of the square target matrix A for some 
    %   q <= k. This implements Algorithm 7. 
    %
    %   [U, D] = S.LowRankPSDApprox() returns the factors U and D that
    %   form the rank-q positive semi-definite approximation Aout = U*D*U'   
    %   for some q <= k. Here, U  is the (m x q) dimensional factor with  
    %   orthonormal columns, and D is of (q x q) dimensions and 
    %   non-negative diagonal matrix. This implements Algorithm 7.
    %
    %   Aout = S.FixedRankSymApprox(r) returns the rank-r conjugate  
    %   symmetric approximation of the square target matrix A. This
    %   implements Algorithm 8. 
    %
    %   [U, D] = S.FixedRankSymApprox(r) returns the factors U and S that
    %   form the rank-r conjugate symmetric approximation Aout = U*D*U'. 
    %   Here, U  is the (m x q) dimensional factor with orthonormal 
    %   columns, and D is of (q x q) dimensions and conjugate symmetric. 
    %   This implements Algorithm 8.
    %
    %   Aout = S.FixedRankPSDApprox(r) returns the rank-r positive  
    %   semi-definite approximation of the square target matrix A. This 
    %   implements Algorithm 9. 
    %
    %   [U, D] = S.FixedRankPSDApprox(r) returns the factors U and D that
    %   form the rank-r positive semi-definite approximation Aout = U*D*U'. 
    %   Here, U  is the (m x q) dimensional factor with orthonormal
    %   columns, and D is of (q x q) dimensions and non-negative diagonal 
    %   matrix. This implements Algorithm 9.
    %
    %   Awoo = S.WoodruffLowRankApprox() returns the rank-q approximation  
    %   of the target matrix A for some q <= k. This implements Woodruff's  
    %   low rank approximation. This is algebraically the same as 
    %   Algorithms 3 and 4, see remark 4.1 in [TYUC2017]
    %
    %   Awoo = S.WoodruffFixedRankApprox(r) the rank-r approximation  
    %   of the target matrix A. This implements Woodruff's fixed rank
    %   approximation formuluation (7.2) in [TYUC2017].
    %
    %   [Q, Z] = S.WoodruffFixedRankApprox(r) returns the factors Q and Z
    %   that forms Awoo = Q*Z, where Awoo is the output of the Woodruff's 
    %   fixed rank approximation formuluation (7.2) in [TYUC2017].
    %
    %   Acemmp = S.CemmpFixedRankApprox(r) the rank-r approximation  
    %   of the target matrix A. This implements Cohen et al.'s fixed rank
    %   approximation formuluation (7.3) in [TYUC2017].
	%
    %   [V, Z] = S.CemmpFixedRankApprox(r) returns the factors V and Z
    %   that forms Acemmp = V*Z, where Acemmp is the output of the Cohen 
    %   et al.'s fixed rank approximation formuluation (7.3) in [TYUC2017].
	%
    %See our reference paper [TYUC2017] for the detailed explanation of the
    %sketching procedure and the arithmetic, communication and storage
    %costs.
    %
	%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
	%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
    %
    %Coded by: Alp Yurtsever
    %Ecole Polytechnique Federale de Lausanne, Switzerland.
    %Laboratory for Information and Inference Systems, LIONS.
    %contact: alp.yurtsever@epfl.ch
    %Created: June 28, 2016
    %Last modified: August 27, 2017
    %
    %PRACTICALSKETCHING-v1.0
    %Copyright (C) 2017 Laboratory for Information and Inference Systems
    %(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
    %
    %This code is a part of PRACTICALSKETCHING toolbox.
    %Please read COPYRIGHT before using this file.

    %% properties
    properties
        Field     % 'real' or 'complex' field (default, 'complex')
        Omega     % (n x k) dimensional test matrix for the range of A
        Psi       % (l x m) dimensional test matrix for the corange of A
        Y         % (m x k) dimensional range sketch
        W         % (l x n) dimensional corange sketch
    end
    
    %% methods
    methods
        % Constructor
        % function obj = Sketch(A, k, l, Field, Orthogonalization)
        function obj = Sketch(A, k, l, varargin)
            % process variable length inputs
            % 1. Field, 2. Orthogonalization
            narginchk(3,5);
            if nargin == 4
                obj.Field  = lower(varargin{1});
            else
                obj.Field = 'complex';
            end
            if nargin == 5
                Orthogonalization  = varargin{2};
            else
                Orthogonalization = 0;
            end
            % construct the sketch
            [m, n] = size(A);
            if (m + n) == 3
                m = A(1);
                n = A(2);
                A = sparse(m,n);
            end
            if strcmp(obj.Field,'real')
                obj.Omega = randn(n,k);
                obj.Psi   = randn(l,m);
            elseif strcmp(obj.Field,'complex')
                obj.Omega = randn(n,k) + 1i*randn(n,k);
                obj.Psi   = randn(l,m) + 1i*randn(l,m);
            else
                error('Field should be ''real'' or ''complex''.');
            end
            if Orthogonalization
                obj.Omega = orth(obj.Omega);
                obj.Psi   = orth(obj.Psi')';
            end
            % initialize the sketch
            obj.Y = A*obj.Omega;
            obj.W = obj.Psi*A;
        end
        
        % Other methods
        obj = LinearUpdate(obj, varargin)           % Algorithm 2 in [MR]
        [Q, X] = SimpleLowRankApprox(obj)           % Algorithm 3
        [Q, X] = LowRankApprox(obj)                 % Algorithm 4
        [U, S] = LowRankSymApprox(obj)              % Algorithm 5
        [U, D] = LowRankPSDApprox(obj)              % Algorithm 6
        [Q, Sigma, V] = FixedRankApprox(obj, r)     % Algorithm 7
        [U, D] = FixedRankSymApprox(obj, r)         % Algorithm 8
        [U, D] = FixedRankPSDApprox(obj, r)         % Algorithm 9
        [Awoo] = WoodruffLowRankApprox(obj)         % Woodruff's low rank approx (4.9) (algebraically equivalent to Algorithm 4, see remark 4.1 in [TYUC2017])
        [Q, Z] = WoodruffFixedRankApprox(obj, r)    % Woodruff's fixed rank approx (7.2) in [TYUC2017]
        [V, Z] = CemmpFixedRankApprox(obj, r)       % Cohen et al.'s fixed rank approx (7.3) in [TYUC2017]
    end
end

% TBD: Add other distributions for the test matrices!
