function [ Q, Z ] = WoodruffFixedRankApprox( obj, r )
%WOODRUFFFIXEDRANKAPPROX Single-View Fixed-Rank Approximation by Woodruff
%This function implements the method [Woo2014, Thm.4.3, display 2] proposed 
%by Woodruff, which takes the form (7.2) in the main reference [TYUC2017].
%   Require: Target rank (r <= k)
%   Ensure:  Returns the low rank reconstrution Awoo of the sketched
%            matrix.
%
%   Awoo = S.WoodruffFixedRankApprox(r) the rank-r approximation
%   of the target matrix A. This implements Woodruff's fixed rank
%   approximation formuluation (5.6) in [TYUC2017].
%
%   [Q, Z] = S.WoodruffFixedRankApprox(r) returns the factors Q and Z
%   that forms Awoo = Q*Z, where Awoo is the output of the Woodruff's
%   fixed rank approximation formuluation (7.2) in [TYUC2017].
%
%See our reference paper [TYUC2017] for the detailed explanation of the 
%sketching procedure and the arithmetic, communication and storage costs.
%
%[TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
%Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%[Woo2014] D.P. Woodruff, Sketching as a Tool for Numerical Linear Algebra,
%Found. Trends Theor. Comput. Sci. 10 (2014)
%
%Coded by: Alp Yurtsever
%Ecole Polytechnique Federale de Lausanne, Switzerland.
%Laboratory for Information and Inference Systems, LIONS.
%contact: alp.yurtsever@epfl.ch
%Created: July 05, 2016
%Last modified: September 2, 2017
%
%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox.
%Please read COPYRIGHT before using this file.

if r > min(size(obj.Omega, 1), size(obj.Psi,2))
    error('Target rank ''r'' cannot be greater than matrix dimensions.')
end

% [U,~] = qr(obj.Psi*obj.Y,0);
% [Uwoo, Sigma, Vwoo] = svds(U'*obj.W,r);
% Awoo = obj.Y*(pinv(obj.Psi*obj.Y)*U*(Uwoo*Sigma*Vwoo'));

[Q, ~] = qr(obj.Y,0);     
[U, T] = qr(obj.Psi*Q,0);   
X = U'*obj.W;
if min(size(X)) > 5000
    [Uwoo, Sigma, Vwoo] = svds(X,r);
else
    [Uwoo, Sigma, Vwoo] = svd(X,'econ');
    Uwoo = Uwoo(:,1:r);
    Sigma = Sigma(1:r,1:r);
    Vwoo = Vwoo(:,1:r);
end
Z = T\(Uwoo*(Sigma*Vwoo'));

if nargout == 1; Q = Q*Z; end

end

