%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox. 
%Please read COPYRIGHT before using this file.
%
%Our primary goal is to obtain the best possible reconstruction of an input
%matrix using the smallest sketch possible. For standard normal test
%matrices, the storage cost of the sketch is directly proportional to the
%sum T = (k + l) of the sketch size parameters. In this experiments, we
%investigate the best way to apportion k and l when we fix the sketch size
%T and the target rank r.
%
%For different values of (T = k+l), we sweep over all possible (k,l) pairs,  
%and record the approximation error, norm(A - Aapprox,'fro').
%
%Test_generic_Type1 and Test_generic_Type2 files are different in the way
%input matrix A is applied. Type1 uses a dense matrix A. Type 2 computes
%the sketch of a low-rank matrix A=Hf*Hb', directly from its factors Hf
%and Hb.
%
%In addition to the algorithms we recommend, our numerical study comprises 
%other methods that have appeared in the literature. See Section 7.3 for  
%the details about these algorithms. 
%
%   See our reference paper [TYUC2017] for the detailed explanation of the
%   sketching procedure and the arithmetic and storage costs.
%
%   [BWZ2016] C. Boutsidis, D. Woodruff and P. Zhong, Optimal principal 
%   component analysis in distributed and streaming models,2016.
%
%   [TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
%   Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%   Coded by: Alp Yurtsever
%   Ecole Polytechnique Federale de Lausanne, Switzerland.
%   Laboratory for Information and Inference Systems, LIONS.
%   contact: alp.yurtsever@epfl.ch
%   Created: August 22, 2016
%   Last modified: August 27, 2017

%% Beginning of the test

T = unique([r*2, round((2*r+3) * 2.^(0:.5:4.5))]);
ErrSpect = (n <= 2e3);

for Iter = 1:numel(T)
    
    k = r;
    l = T(Iter) - k;
    StorageCost = 2*T(Iter)*n;
    
    %% Preallocation
    sd.k{Iter} = [];
    sd.l{Iter} = [];
    sd.t{Iter} = [];
    sd.s{Iter} = [];
    sd.ErrA7Fro{Iter} = [];
    sd.ErrA8Fro{Iter} = [];
    sd.ErrA9Fro{Iter} = [];
    sd.ErrWooFro{Iter} = [];
    sd.ErrCemmpFro{Iter} = [];
    sd.ErrBwzFro{Iter} = [];
    if ErrSpect
        sd.ErrA7Spect{Iter} = [];
        sd.ErrA8Spect{Iter} = [];
        sd.ErrA9Spect{Iter} = [];
        sd.ErrWooSpect{Iter} = [];
        sd.ErrCemmpSpect{Iter} = [];
        sd.ErrBwzSpect{Iter} = [];
    end
    
    while r <= l
        %% Generate sketch
        mySketch = Sketch(A, k, l, field);
        
        if (k < l) && (r < k)
            %% Our Single-View Low-Rank Approximation (Fixed rank reconstruction)
            A7 = mySketch.FixedRankApprox(r);
            ErrA7Fro  = norm(A7 - A,'fro');
            if ErrSpect
                ErrA7Spect  = norm(A7 - A);
            end
            %% Our Single-View Low-Rank Approximation (Fixed rank symmetric reconstruction)
            A8 = mySketch.FixedRankSymApprox(r);
            ErrA8Fro  = norm(A8 - A,'fro');
            if ErrSpect
                ErrA8Spect  = norm(A8 - A);
            end
            %% Our Single-View Low-Rank Approximation (Fixed rank psd symmetric reconstruction)
            A9 = mySketch.FixedRankPSDApprox(r);
            ErrA9Fro  = norm(A9 - A,'fro');
            if ErrSpect
                ErrA9Spect  = norm(A9 - A);
            end
        else
            ErrA7Fro = nan;
            ErrA8Fro = nan;
            ErrA9Fro = nan;
            if ErrSpect
                ErrA7Spect  = nan;
                ErrA8Spect  = nan;
                ErrA9Spect  = nan;
            end
        end
        %% Woodruff's Single-View Fixed-Rank Approximation
        Awoo = mySketch.WoodruffFixedRankApprox(r);
        ErrWooFro  = norm(Awoo - A,'fro');
        if ErrSpect
            ErrWooSpect = norm(Awoo - A);
        end

        %% Cohen et al.'s Single-View Fixed-Rank Approximation
        Acemmp = mySketch.CemmpFixedRankApprox(r);
        ErrCemmpFro  = norm(Acemmp - A,'fro');
        if ErrSpect
            ErrCemmpSpect = norm(Acemmp - A);
        end
        
        %% Print intermediate results
        fprintf('k+l = %d, k = %d, l = %d, A7RelErr = %f, A8RelErr = %f, A9RelErr = %f, WooRelErr = %f, CemmpRelErr = %f  ... \n', ...
            T(Iter), k, l, ErrA7Fro./ErrBestFro-1, ErrA8Fro./ErrBestFro-1, ErrA9Fro./ErrBestFro-1, ErrWooFro./ErrBestFro-1, ErrCemmpFro./ErrBestFro-1);
        
        %% Save data
        sd.k{Iter}(end+1) = k;
        sd.l{Iter}(end+1) = l;
        sd.ErrA7Fro{Iter}(end+1) = ErrA7Fro;
        sd.ErrA8Fro{Iter}(end+1) = ErrA8Fro;
        sd.ErrA9Fro{Iter}(end+1) = ErrA9Fro;
        sd.ErrWooFro{Iter}(end+1) = ErrWooFro;
        sd.ErrCemmpFro{Iter}(end+1) = ErrCemmpFro;
        if ErrSpect
            sd.ErrA7Spect{Iter}(end+1) = ErrA7Spect;
            sd.ErrA8Spect{Iter}(end+1) = ErrA8Spect;
            sd.ErrA9Spect{Iter}(end+1) = ErrA9Spect;
            sd.ErrWooSpect{Iter}(end+1) = ErrWooSpect;
            sd.ErrCemmpSpect{Iter}(end+1) = ErrCemmpSpect;
        end
        %% increase k, decrease l, keep k+l constant
        k = k+1;
        l = l-1;
    end
    
    %% [BWZ2016]
    % t corresponds to k in the paper. Used as "t" in the code to avoid
    % notational conflict with the k used in the first part
    t = r;
    s = inf;
    while t <= s
        %It is trickier to include the [BWZ2016] (presented in [TYUC2016] eqn (7.6)). 
        %For a given T, we obtain the oracle performance by minimizing the empirical 
        %approximation error over pairs (t, s) for which the storage cost of the sketch 
        %matches the cost of the simple sketch. 
        %That is, (2t + 1)2n + s(s + 2) ~ 2Tn 
        %   ==>  s^2 + 2s - 2Tn + 4nt + 2n = 0
        %s is approximated by solving this equation and rounding to the closest integer. 
        
        % % ignoring s^2 term
        % s = round(sqrt(StorageCost - 4*t*n - 2*n));
        
        % % taking s^2 term into account
        s = round(sqrt(StorageCost + 1 - 4*t*n - 2*n) - 1);
        s = min(n,s);
        
        if (r <= t) && (t <= s)
            %% Create the optimal sketch
            myOptSketch = OptSketch(A, t, s, field);
            
            %% Updahyay's Single-View Fixed-Rank Approximation
            Abwz = myOptSketch.BwzFixedRankApprox(r);
            ErrBwzFro  = norm(Abwz - A,'fro');
            if ErrSpect
                ErrBwzSpect  = norm(Abwz - A);
            end
        else
            ErrBwzFro = nan;
            if ErrSpect
                ErrBwzSpect = nan;
            end
        end
        
        %% Compute the errors
        % Find (A - Aour) using thinSVD
        fprintf('k+l = %d, t = %d, v = %d, RelErrBwzFro = %f ... \n', ...
            T(Iter), t, s, ErrBwzFro./ErrBestFro-1 );
        
        %% Save data
        sd.t{Iter}(end+1) = t;
        sd.s{Iter}(end+1) = s;
        sd.ErrBwzFro{Iter}(end+1) = ErrBwzFro;
        if ErrSpect
            sd.ErrBwzSpect{Iter}(end+1) = ErrBwzSpect;
        end
        %% increase t
        t = t+1;
    end
    
end

%% Save results

% Variables to be saved
sd.ErrBestFro = ErrBestFro;
sd.ErrBestSpect = ErrBestSpect;
sd.info.n = n;
sd.info.r = r;
sd.info.field = field;
sd.info.T = T;
sd.ID = [datestr(now,30),num2str(randi([100,999]))];

