%PRACTICALSKETCHING-v1.0
%Copyright (C) 2017 Laboratory for Information and Inference Systems
%(LIONS), Ecole Polytechnique Federale de Lausanne, Switzerland.
%
%This code is a part of PRACTICALSKETCHING toolbox. 
%Please read COPYRIGHT before using this file.
%
%Our primary goal is to obtain the best possible reconstruction of an input
%matrix using the smallest sketch possible. For standard normal test
%matrices, the storage cost of the sketch is directly proportional to the
%sum T = (k + l) of the sketch size parameters. In this experiments, we
%investigate the best way to apportion k and l when we fix the sketch size
%T and the target rank r.
%
%For different values of (T = k+l), we sweep over all possible (k,l) pairs,  
%and record the approximation error, norm(A - Aapprox,'fro').
%
%Test_generic_Type1 and Test_generic_Type2 files are different in the way
%input matrix A is applied. Type1 uses a dense matrix A. Type 2 computes
%the sketch of a low-rank matrix A=Hf*Hb', directly from its factors Hf
%and Hb.
%
%In addition to the algorithms we recommend, our numerical study comprises 
%other methods that have appeared in the literature. See Section 7.3 for  
%the details about these algorithms. 
%
%   See our reference paper [TYUC2017] for the detailed explanation of the
%   sketching procedure and the arithmetic and storage costs.
%
%   [BWZ2016] C. Boutsidis, D. Woodruff and P. Zhong, Optimal Principal 
%   Component Analysis in Distributed and Streaming Models, 2016.
%
%   [TYUC2017] J.A. Tropp, A. Yurtsever, M. Udell and V. Cevher, Practical  
%   Sketching Algorithms for Low-Rank Matrix Approximation, SIMAX, 2017.
%
%   Coded by: Alp Yurtsever
%   Ecole Polytechnique Federale de Lausanne, Switzerland.
%   Laboratory for Information and Inference Systems, LIONS.
%   contact: alp.yurtsever@epfl.ch
%   Created: August 22, 2016
%   Last modified: August 27, 2017

%% Beginning of the test

% sweep region of T=(k+l)
T = unique([r*2, round((2*r+3) * 2.^(0:.5:4))]);  

for Iter = 1:numel(T)
    
    k = r;
    l = T(Iter) - k;
    StorageCost = 2*T(Iter)*n;
        
    %% Preallocation
    sd.k{Iter} = [];
    sd.l{Iter} = [];
    sd.t{Iter} = [];
    sd.s{Iter} = [];
    sd.ErrA7Fro{Iter} = [];
    sd.ErrA7Spect{Iter} = [];
    sd.ErrA8Fro{Iter} = [];
    sd.ErrA8Spect{Iter} = [];
    sd.ErrA9Fro{Iter} = [];
    sd.ErrA9Spect{Iter} = [];
    sd.ErrWooFro{Iter} = [];
    sd.ErrWooSpect{Iter} = [];
    sd.ErrCemmpFro{Iter} = [];
    sd.ErrCemmpSpect{Iter} = [];
    sd.ErrBwzFro{Iter} = [];
    sd.ErrBwzSpect{Iter} = [];
    
    while r <= l
        %% Our Single-View Fixed-Rank Approximation with structures
        mySketch = Sketch([n, n], k, l, field, 0);
        mySketch = mySketch.LinearUpdate(U*Z, U, 0, 1);
        
        if (k < l) && (r < k)            
            %% Our Single-View Low-Rank Approximation (Fixed rank reconstruction)
            [Q7, Sigma7, V7] = mySketch.FixedRankApprox(r);
            [~,ZZ] = updateThinSVD(U,Z,U,Q7*(-Sigma7),V7);
            ErrA7Fro  = norm(diag(ZZ));
            ErrA7Spect  = norm(diag(ZZ),inf);
            %% Our Single-View Low-Rank Approximation (Fixed rank symmetric reconstruction)
            [Q8, Z8] = mySketch.FixedRankSymApprox(r);
            [~,ZZ] = updateThinSVD(U,Z,U,Q8*(-Z8),Q8);
            ErrA8Fro  = norm(diag(ZZ));
            ErrA8Spect  = norm(diag(ZZ),inf);
            %% Our Single-View Fixed rank PSD Approximation (Alg.9)
            [Q9, Z9] = mySketch.FixedRankPSDApprox(r);
            [~,ZZ] = updateThinSVD(U,Z,U,Q9*(-Z9),Q9);
            ErrA9Fro  = norm(diag(ZZ));
            ErrA9Spect  = norm(diag(ZZ),inf);
        else
            ErrA7Fro = nan;
            ErrA7Spect = nan;
            ErrA8Fro = nan;
            ErrA8Spect = nan;
            ErrA9Fro = nan;
            ErrA9Spect = nan;
        end
        
        %% Woodruff's Single-View Fixed-Rank Approximation
        [Qwoo,Zwoo] = mySketch.WoodruffFixedRankApprox(r);
        [~,ZZ] = updateThinSVD(U,Z,U,-Qwoo,Zwoo');
        ErrWooFro  = norm(diag(ZZ));
        ErrWooSpect  = norm(diag(ZZ),inf);

        %% Cohen et al.'s Single-View Fixed-Rank Approximation
        [Qcemmp,Zcemmp] = mySketch.CemmpFixedRankApprox(r);
        [~,ZZ] = updateThinSVD(U,Z,U,-Qcemmp,Zcemmp');
        ErrCemmpFro  = norm(diag(ZZ));
        ErrCemmpSpect  = norm(diag(ZZ),inf);
        
        %% Compute the errors
        % Find (A - Aour) using thinSVD
        fprintf('k+l = %d, k = %d, l = %d, A7RelErr = %f, A9RelErr = %f, WooRelErr = %f  ... \n', ... 
            T(Iter), k, l, ErrA7Fro./ErrBestFro-1, ErrA9Fro./ErrBestFro-1, ErrWooFro./ErrBestFro-1);
        
        %% Save data
        sd.k{Iter}(end+1) = k;
        sd.l{Iter}(end+1) = l;
        sd.ErrA7Fro{Iter}(end+1) = ErrA7Fro;
        sd.ErrA7Spect{Iter}(end+1) = ErrA7Spect;
        sd.ErrA8Fro{Iter}(end+1) = ErrA8Fro;
        sd.ErrA8Spect{Iter}(end+1) = ErrA8Spect;
        sd.ErrA9Fro{Iter}(end+1) = ErrA9Fro;
        sd.ErrA9Spect{Iter}(end+1) = ErrA9Spect;
        sd.ErrWooFro{Iter}(end+1) = ErrWooFro;
        sd.ErrWooSpect{Iter}(end+1) = ErrWooSpect;
        sd.ErrCemmpFro{Iter}(end+1) = ErrCemmpFro;
        sd.ErrCemmpSpect{Iter}(end+1) = ErrCemmpSpect;

        %% increase k, decrease l, keep T=k+l constant
        k = k+1;
        l = l-1;
    end
    
    %% [UPA16]
    % t corresponds to k in the paper. Used as "t" in the code to avoid
    % notational conflict with the k used in the first part
    t = r;   
    s = inf;
    while t <= s
        %%compute s for given t
        %It is trickier to include the [BWZ2016] (presented in [TYUC2016] eqn (7.6)). 
        %For a given T, we obtain the oracle performance by minimizing the empirical 
        %approximation error over pairs (t, s) for which the storage cost of the sketch 
        %matches the cost of the simple sketch. 
        %That is, (2t + 1)2n + s(s + 2) ~ 2Tn 
        %   ==>  s^2 + 2s - 2Tn + 4nt + 2n = 0
        %s is approximated by solving this equation and rounding to the closest integer. 
        
        % % ignoring s^2 term
        % s = round(sqrt(StorageCost - 4*t*n - 2*n));
        
        % % taking s^2 term into account
        s = round(sqrt(StorageCost + 1 - 4*t*n - 2*n) - 1);
        
        s = min(n,s);
        
        if (r <= t) && (t <= s)
            %% Create the optimal sketch
            myOptSketch = OptSketch([n, n], t, s, field, 0);
            myOptSketch = myOptSketch.LinearUpdate(U*Z, U, 0, 1);
            
            %% Updahyay's Single-View Fixed-Rank Approximation
            [Qbwz, SigmaBwz, Vbwz] = myOptSketch.BwzFixedRankApprox(r);
            [~,ZZ] = updateThinSVD(U,Z,U,Qbwz*(-SigmaBwz),Vbwz);
            ErrBwzFro  = norm(diag(ZZ));
            ErrBwzSpect  = norm(diag(ZZ),inf);
            
        else
            ErrBwzFro = nan;
            ErrBwzSpect = nan;
        end
        
        %% Compute the errors
        % Find (A - Aour) using thinSVD
        fprintf('k+l = %d, t = %d, v = %d, RelErrBwzFro = %f \n', ...
            T(Iter), t, s, ErrBwzFro./ErrBestFro-1 );
        
        %% Save data
        sd.t{Iter}(end+1) = t;
        sd.s{Iter}(end+1) = s;
        sd.ErrBwzFro{Iter}(end+1) = ErrBwzFro;
        sd.ErrBwzSpect{Iter}(end+1) = ErrBwzSpect;
        %% increase t
        t = t+1;
    end
    
end

%% Variables to be saved
sd.ErrBestFro = ErrBestFro;
sd.ErrBestSpect = ErrBestSpect;
sd.info.n = n;
sd.info.r = r;
sd.info.field = field;
sd.info.T = T;
sd.ID = [datestr(now,30),num2str(randi([100,999]))];

